import discord
import asyncio
import random
import json
import re 
from T3SF import utils

class Discord(object):
	def __init__(self, bot):
		self.bot = bot

	async def InboxesAuto(self, T3SF_instance):
		mensaje_inboxes = ""
		image_example = "https://i.ibb.co/NCrPD3Y/discord-exp.png"

		if T3SF_instance.fetch_inboxes == True:
			if T3SF_instance._ctx == None:
				# Exercise started from the GUI
				guild = self.bot.get_guild(int(T3SF_instance.guild_id))
				channels = guild.text_channels

				for channel in channels:
					if "chat" in channel.name:
						accuracy = utils.similar(str(channel.name).lower(),"gm-chat")
						if accuracy >= 0.8:
							inbox = channel.id
							T3SF_instance._ctx = self.bot.get_channel(channel.id)

			T3SF_instance.response_auto = await self.SendMessage(T3SF_instance=T3SF_instance, title="⚙️ Fetching inboxes...", description=f"Please wait while we fetch all the inboxes in this server!", color="BLUE")

			channels_itinerator = 0
			player_itinerator = 0
			regex = ""

			try:
				channels = T3SF_instance._ctx.message.guild.channels
				categories = T3SF_instance._ctx.message.guild.categories
				started_from_gui = False

			except Exception:
				channels = T3SF_instance._ctx.guild.channels
				categories = T3SF_instance._ctx.guild.categories
				started_from_gui = True

			while regex == "" and len(T3SF_instance.players_list) > player_itinerator:
				for category in categories:
					if channels_itinerator == 0:
						past_channel = category.name
						pass
					
					match_channel = utils.similar(str(category).lower(), str(T3SF_instance.players_list[0]).lower())
					
					if match_channel >= 0.4:
						for character in past_channel:
							if character in category.name:
								regex += character
							else:
								break
					else:
						player_itinerator += 1
						channels_itinerator += 1
						past_channel = category.name
						pass

			await T3SF_instance.response_auto.edit(embed=discord.Embed(
			 title = "ℹ️ Regex detected!",
			 description = f"Please confirm if the regex detected for the channels, is correct so we can get the inboxes!\n\nExample:\nGroup - Legal\nThe regex should be `Group -`\n\nDetected regex: `{regex}`",
			 color = 0x77B255).set_image(url=image_example).set_footer(text="Please answer with [Yes(Y)/No(N)]"))

			def check_regex_channels(msg):
				if started_from_gui:
					return msg.content.lower() in ["y", "yes", "n", "no"]

				else:
					return msg.author == T3SF_instance._ctx.author and msg.channel == T3SF_instance._ctx.channel and msg.content.lower() in ["y", "yes", "n", "no"]

			try:
				msg = await self.bot.wait_for("message", check=check_regex_channels, timeout=50)

				if msg.content.lower() in ["y", "yes"]:
					await self.EditMessage(T3SF_instance=T3SF_instance, style="custom", variable="T3SF_instance.response_auto", color="GREEN", title = "✨ Regex detected succesfully! ✨", description = f"Thanks for confirming the regex detected for the channels (I'm going to tell my creator he is so good coding :D ), we are going to use `{regex}` to match the inboxes")

				elif msg.content.lower() in ["n", "no"]:
					await T3SF_instance.response_auto.edit(embed=discord.Embed(
					 title = "ℹ️ Regex needed!",
					 description = "Got it!\n Unluckily, but here we go...\nPlease send me the regex for the channels, so we can get the inboxes!\n\nExample:\nGroup - Legal\nThe regex should be `Group -`",
					 color = discord.Colour.red()).set_image(url=image_example).set_footer(text="Please answer with the desired regex. EG: `Groups -`"))

					def get_regex_channels(msg_regex_user):
						# return msg_regex_user.author == T3SF_instance._ctx.author and msg_regex_user.channel == T3SF_instance._ctx.channel and 

						if started_from_gui:
							return msg_regex_user.content != ""

						else:
							return msg_regex_user.author == T3SF_instance._ctx.author and msg_regex_user.channel == T3SF_instance._ctx.channel and msg_regex_user.content != ""
					
					msg_regex_user = await self.bot.wait_for("message", check=get_regex_channels, timeout=50)

					if msg_regex_user.content != "":
						regex = msg_regex_user
						await self.EditMessage(T3SF_instance=T3SF_instance, style="custom", variable="T3SF_instance.response_auto", color="GREEN", title="✅ Regex accepted!", description=f"Thanks for confirming the regex for the channels, we are going to use `{msg_regex_user.content}` to match the inboxes!")

			except asyncio.TimeoutError:
				await self.EditMessage(T3SF_instance=T3SF_instance, style="custom", variable="T3SF_instance.response_auto", color="RED", title = ":x: Sorry, you didn't reply on time", description = "Please start the process again.")
				raise RuntimeError("We didn't detect any regex, time's up.")

			for player in T3SF_instance.players_list:
				for channel in channels:
					category = channel.category
					if "inbox" in channel.name :
						accuracy = utils.similar(re.sub(f"({regex})", "",str(category)).lower(),str(player).lower())
						if accuracy >= 0.4:
							T3SF_instance.inboxes_all[player] = channel.id
			
			json.dump(T3SF_instance.inboxes_all,open(f"inboxes_{T3SF_instance.platform}.json", "w"))

			for player in T3SF_instance.inboxes_all:
				mensaje_inboxes += f"**Inbox** {player}[{T3SF_instance.inboxes_all[player]}]\n"

			await self.EditMessage(T3SF_instance=T3SF_instance, style="custom", variable="T3SF_instance.response_auto", color="GREEN", title=f"📩  Inboxes fetched! [{len(T3SF_instance.inboxes_all)}]", description=mensaje_inboxes)

		return True

	async def InjectHandler(self, T3SF_instance):
		all_data = f'Date: {T3SF_instance._inject["Date"]}\n\n{T3SF_instance._inject["Script"]}'

		player = T3SF_instance._inject['Player']

		inbox = self.bot.get_channel(T3SF_instance.inboxes_all[player])

		embed = discord.Embed(title = T3SF_instance._inject['Subject'], description = all_data, color = discord.Colour.blue())
		
		if "Photo" in T3SF_instance._inject and T3SF_instance._inject['Photo'] != '':
			embed.set_image(url=T3SF_instance._inject['Photo'])
		
		if "Profile" in T3SF_instance._inject and T3SF_instance._inject['Profile'] != '':
			profile_pic = T3SF_instance._inject['Profile']
		else:
			profile_pic = random.choice([
				"https://ssl.gstatic.com/ui/v1/icons/mail/profile_mask2.png",
				"https://lh3.googleusercontent.com/-XdUIqdMkCWA/AAAAAAAAAAI/AAAAAAAAAAA/4252rscbv5M/photo.jpg",
				"https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcSTt8Dg9RL4IGOjsJ2Fr-lXThf-DGM5YgPB6j5rD8tHQ9RLrU-03H4dYeskL01FNajqL_0&usqp=CAU"
			])

		embed.set_author(name=T3SF_instance._inject["From"], icon_url=profile_pic)

		T3SF_instance.response_poll = await inbox.send(embed = embed)

		return embed

	async def PollHandler(self, T3SF_instance):
		T3SF_instance.poll_answered = False

		all_data = T3SF_instance._inject["Script"]

		poll_options = T3SF_instance._inject['Poll'].split('|')
		
		actual_real_time = re.sub("([^0-9])", "", T3SF_instance._inject['Real Time'])[-2:]
		
		next_real_time = re.sub("([^0-9])", "", T3SF_instance.data[int(T3SF_instance._inject['#'])]['Real Time'])[-2:]

		diff = int(next_real_time) - int(actual_real_time)
		if diff < 0:
			diff_no_real = int(actual_real_time) - int(next_real_time)
			diff = 60 - diff_no_real 

		diff_secs = diff * 60

		view = discord.ui.View()
		view.add_item(discord.ui.Button(style=discord.ButtonStyle.primary,label=poll_options[0], custom_id="poll|"+poll_options[0]))
		view.add_item(discord.ui.Button(style=discord.ButtonStyle.primary,label=poll_options[1], custom_id="poll|"+poll_options[1]))

		all_data = all_data+ f"\n\nYou have {diff} minute(s) to answer this poll!"

		player = T3SF_instance._inject['Player']

		inbox = self.bot.get_channel(T3SF_instance.inboxes_all[player])

		embed = discord.Embed(title = T3SF_instance._inject['Subject'], description = all_data, color = discord.Colour.yellow())

		if "Photo" in T3SF_instance._inject and T3SF_instance._inject['Photo'] != '':
			embed.set_image(url=T3SF_instance._inject['Photo'])

		T3SF_instance.response_poll = await inbox.send(embed = embed, view=view)

		return embed

	async def PollAnswerHandler(self, T3SF_instance, interaction=None):
		if "poll" in interaction.data['custom_id']:
			poll_msg_og = interaction.message.embeds.copy()[0]

			title =  poll_msg_og.title

			poll_msg = poll_msg_og.description
			poll_msg = poll_msg[: poll_msg.rfind('\n')]

			action_user = interaction.user

			selected_option = interaction.data['custom_id'].split('|')[1]
			description = f'{poll_msg}\n\n@{action_user} selected: {selected_option}'               

			T3SF_instance.poll_answered = True
			T3SF_instance.response_poll = await interaction.response.edit_message(embed=discord.Embed(colour=discord.Colour.green(), title=title, description=description),view=None)
			await T3SF_instance.NotifyGameMasters(type_info="poll_answered", data={'msg_poll':poll_msg,'answer':selected_option,'user':action_user})
			return True
		else:
			pass

	async def SendMessage(self, T3SF_instance, color="CYAN", title:str=None, description:str=None, view=None, unique=False):
		colors = {'BLUE' : discord.Colour.dark_blue(), 'RED' : discord.Colour.red(), 'CYAN' : discord.Colour.blue(), 'GREEN' : discord.Colour.green(), 'YELLOW' : discord.Colour.yellow()}

		if unique == True:
			T3SF_instance.gm_poll_msg = await T3SF_instance._ctx.send(embed=discord.Embed(color=colors[color], title = title, description = description), view=view)
			return T3SF_instance.gm_poll_msg
		
		else:
			T3SF_instance.response = await T3SF_instance._ctx.send(embed=discord.Embed(color=colors[color], title = title, description = description), view=view)
			return T3SF_instance.response

	async def EditMessage(self, T3SF_instance=None, color="CYAN", title:str=None, description:str=None, view=None, style="simple", variable=None):

		if style == "simple":
			T3SF_instance.response = T3SF_instance.response.edit(embed=discord.Embed(color=colors[color], title = title, description = description), view=view)
			return T3SF_instance.response
		else:
			colors = {'BLUE' : discord.Colour.dark_blue(), 'RED' : discord.Colour.red(), 'CYAN' : discord.Colour.blue(), 'GREEN' : discord.Colour.green(), 'YELLOW' : discord.Colour.yellow()}
			variable = eval(variable)
			await variable.edit(embed=discord.Embed(color=colors[color], title = title, description = description), view=view)


from discord.ext import commands
from dotenv import load_dotenv
import discord
import asyncio
import re,os
import T3SF 

load_dotenv()

TOKEN = os.environ['DISCORD_TOKEN']

intents = discord.Intents.default()
intents.message_content = True
bot_discord = commands.Bot(command_prefix='!', intents=intents)

# Defining global variables and events
T3SF_instance = None
start_incidents_gui = asyncio.Event()
create_env = asyncio.Event()
config_MSEL = None

class create_bot():
	def __init__(self, MSEL):
		global config_MSEL
		config_MSEL = MSEL
		self.define_commands()

	def define_commands(self):
		global bot_discord
		
		@bot_discord.event
		async def on_ready():
			"""
			Print some infomation about the bot when it connects succesfully.
			"""
			print(f'\n * Discord Bot - {bot_discord.user} is ready!')
			T3SF.T3SF_Logger.emit(message=f'Discord Bot - {bot_discord.user} is ready!', message_type="DEBUG")

		@bot_discord.event
		async def on_interaction(interaction):
			await T3SF.T3SF.PollAnswerHandler(self=T3SF_instance, payload=interaction)

		@bot_discord.command(name='ping', help='Responds with pong to know if the bot is up! Usage -> !ping')
		async def ping_command(ctx):
			"""
			Retrieves the !ping command and replies with the latency.
			"""
			description = f"""
			**Pong!**

			:stopwatch: `{round(bot_discord.latency*1000)}ms`
			"""
			response = await ctx.send(embed=discord.Embed(colour=discord.Colour.blue(), description=description))

		@bot_discord.command(name="start", help='Starts the Incidents Game. Usage -> !start')
		@commands.has_role("Game Master")
		async def start_command(ctx, *, query=None):
			"""
			Retrieves the !start command and starts to fetch and send the incidents.
			"""
			try:
				global T3SF_instance
				T3SF_instance = T3SF.T3SF(platform="discord", bot=bot_discord)
				await T3SF_instance.ProcessIncidents(MSEL = config_MSEL, function_type = "start", ctx=ctx)

			except Exception as e:
				print("ERROR - Start function")
				print(e)
				raise

		@bot_discord.event
		async def on_command_error(ctx, error):
		    """
		    Inform if an error with the commands ocurred.
		    """
		    if isinstance(error, commands.errors.CheckFailure):
		        await ctx.send(embed=discord.Embed(colour=discord.Colour.red(), title="Error 401", description='You do not have the "Game Master" role to use this command.'))

		    elif isinstance(error, discord.HTTPException):
		        await ctx.send(embed=discord.Embed(colour=discord.Colour.red(), title="Error 5xx", description="We got ratelimited by Discord. Please wait a few seconds and try again."))

async def start_bot():
	T3SF_instance = T3SF.T3SF(platform="discord", bot=bot_discord)
	task1 = asyncio.create_task(bot_discord.start(TOKEN))
	task2 = asyncio.create_task(async_handler_exercise())
	task3 = asyncio.create_task(create_environment_task())

	# wait for the coroutines to finish
	await asyncio.gather(task1, task2, task3)

async def async_handler_exercise():
	global T3SF_instance
	while True:
		try:
			await asyncio.wait_for(start_incidents_gui.wait(), timeout=0.1)
		except asyncio.TimeoutError:
			pass
		else:
			T3SF_instance = T3SF.T3SF(platform="discord", bot=bot_discord)
			T3SF_instance.guild_id = server_id
			asyncio.create_task(T3SF_instance.ProcessIncidents(MSEL = config_MSEL, function_type="start", ctx=None))

			start_incidents_gui.clear()

async def run_async_incidents(server):
	global server_id
	server_id = server
	start_incidents_gui.set()

async def create_environment(server):
	global server_id
	server_id = server
	create_env.set()

async def create_environment_task():
	global T3SF_instance
	while True:
		try:
			await asyncio.wait_for(create_env.wait(), timeout=0.1)
		except asyncio.TimeoutError:
			pass
		else:
			create_env.clear()

			T3SF_instance = T3SF.T3SF(platform="discord", bot=bot_discord)
			areas_msel = T3SF_instance.IncidentsFetcher(MSEL = config_MSEL)

			# Set the GUILD To the user's input server
			guild = bot_discord.get_guild(int(server_id))

			inbox_name = "inbox"
			extra_chnls = ['chat','decision-log']
			players_list_local = areas_msel

			await create_gm_channels(guild=guild)

			for player in players_list_local:

				player_normalized = player.lower().replace(" ", "-")

				role = await create_role_if_not_exists(guild=guild, name=player_normalized)

				if role:
					category_name =  "Group - " + player
					category = await create_category_if_not_exists(guild=guild, name=category_name, private=True, role=role)

					if category:
						await create_channel_if_not_exists(category=category, name=inbox_name)
						T3SF.T3SF_Logger.emit(message=f'Player [{player}] - Channel {inbox_name} created', message_type="DEBUG")

						for channel in extra_chnls:
							await create_channel_if_not_exists(category=category, name=channel)
							T3SF.T3SF_Logger.emit(message=f'Player [{player}] - Channel {channel} created', message_type="DEBUG")
						
						await create_voice_if_not_exists(category=category, name="office")
						T3SF.T3SF_Logger.emit(message=f'Player [{player}] - Voice Channel created', message_type="DEBUG")

			T3SF.T3SF_Logger.emit(message=f"Player's Channels created!", message_type="INFO")

async def create_role_if_not_exists(guild, name):
	try:
		role = discord.utils.get(guild.roles, name=name)

		if role is None:
			general = discord.Permissions(view_channel=True)
			text = discord.Permissions.text()
			voice = discord.Permissions.voice()
			role = await guild.create_role(name=name, permissions=general | text | voice, colour=discord.Colour.red())
			return role
		else:
			return role

	except Exception as e:
		print(e)
		raise

async def create_category_if_not_exists(guild, name, private=False, role=None):
	try:
		category = discord.utils.get(guild.categories, name=name)

		if category is None:
			category = await guild.create_category(name)

			if private:
				await category.set_permissions(guild.default_role, read_messages=False)
				await category.set_permissions(role, read_messages=True)

			return category
		else:
			return category

	except Exception as e:
		print(e)
		raise

async def create_channel_if_not_exists(category, name):
	try:
		channel = discord.utils.get(category.channels, name=name)
		if channel is None:
			channel = await category.create_text_channel(name)
			return channel	
		else:
			return channel
	except Exception as e:
		print(e)
		raise

async def create_voice_if_not_exists(category, name):
	try:
		channel = discord.utils.get(category.channels, name=name)
		if channel is None:
			channel = await category.create_voice_channel(name)
			return channel	
		else:
			return channel
	except Exception as e:
		print(e)
		raise

async def create_gm_channels(guild):

	role = discord.utils.get(guild.roles, name="Game Master")
	
	if role is None:
		all_perms = discord.Permissions.all()
		role = await guild.create_role(name="Game Master", permissions=all_perms, colour=discord.Colour.green())

	category = await create_category_if_not_exists(guild=guild, name="Control Room", private=True, role=role)
	channels = ['gm-chat','logs']

	for channel in channels:
		await create_channel_if_not_exists(category=category, name=channel)
	
	await create_voice_if_not_exists(category=category, name="Game Masters")

	T3SF.T3SF_Logger.emit(message=f'Game Master channels created.', message_type="INFO")

	return True


from collections import Counter
from T3SF import utils
import random
import json
import re 

class Slack(object):
	def __init__(self, app):
		self.app = app

	def Formatter(self=None, title=None, description=None, color="CYAN", image=None, author=None, buttons=None, text_input=None, checkboxes=None):
		colors = {'BLUE' : '#428bca', 'RED' : '#d9534f', 'WHITE' : '#f9f9f9', 'CYAN' : '#5bc0de', 'GREEN' : '#5cb85c', 'ORANGE' : '#ffa700', 'YELLOW' : '#ffff00'}

		fallback_text = ""
		color = colors[color]
		result =[
					{
						"color": color,
						"blocks": []
					}
				]

		if title != None:
			title_list = [
							{
								"type": "header",
								"text": {
									"type": "plain_text",
									"text": title
								}
							},
							{
								"type": "divider"
							}
						]

			result[0]["blocks"] = title_list

		if description != None:
			desc_list = {
							"type": "section",
							"text": {
								"type": "mrkdwn",
								"text": description
							}
						}
			result[0]["blocks"].append(desc_list)

		else:
			description = "Preview not available."
			desc_list = {
				"type": "section",
				"text": {
					"type": "mrkdwn",
					"text": description
				}
			}
			result[0]["blocks"].append(desc_list)

		if image != None:
			fallback_text = "🖼 "
			image_list = {
							"type": "image",
							"title": {
								"type": "plain_text",
								"text": image["name"],
								"emoji": True
							},
							"image_url": image["image_url"],
							"alt_text": "image"
						}
			result[0]["blocks"].append(image_list)

		if author != None:
			author_list = [{
							"type": "context",
							"elements": [
								{
									"type": "image",
									"image_url": author["image_url"],
									"alt_text": author["name"]
								},
								{
									"type": "mrkdwn",
									"text": author["name"],
								},
								{
									"type": "mrkdwn",
									"text": " |  *" + author["date"] + "*",
								}
							]
						  }]
			result[0]["blocks"][1:1] = author_list

		if text_input != None:
			input_list = {
							"block_id": "input_texto",
							"dispatch_action": text_input['dispatch_action'],
							"type": "input",
							"element": {
								"type": "plain_text_input",
								"action_id": text_input["action_id"]
							},
							"label": {
								"type": "plain_text",
								"text": text_input["label"]
							}
						}
			result[0]["blocks"].append(input_list)

		if checkboxes != None:
			chk_boxes_list = {
				"block_id": "checkboxes",
				"type": "input",
				"label": {
					"type": "plain_text",
					"text": checkboxes['title'],
					"emoji": True
				},
				"element": {
					"type": "checkboxes",
					"options": []
				}
			}
			if 'action_id' in checkboxes:
				chk_boxes_list['element']["action_id"] = checkboxes['action_id']

			for checkbox in checkboxes['checkboxes']:
				boxes_list = {
							"text": {
								"type": "plain_text",
								"text": checkbox['text'],
								"emoji": True
							},
							"value": checkbox['value']
							}
				chk_boxes_list['element']['options'].append(boxes_list)

			result[0]["blocks"].append(chk_boxes_list)

		if buttons != None:
			actions_list = {
							"type": "actions",
							"elements": []
						   }
			
			for button in buttons:
				button_list = {
								"type": "button",
								"text": {
									"type": "plain_text",
									"text": button['text']
								},
								"style": button['style'],
								"value": button['value'],
								"action_id": button['action_id']
								}
				actions_list['elements'].append(button_list)

			result[0]["blocks"].append(actions_list)

		result[0]["fallback"] = (fallback_text + description)

		return result

	async def InboxesAuto(self, T3SF_instance, regex=None):
		if regex != None:
			mensaje_inboxes = ""
			for player in T3SF_instance.players_list:
				for channel in T3SF_instance.ch_names_list:
					if regex in channel:
						accuracy = utils.similar(re.sub(f"({regex})", "", str(channel)).lower(), str(player).lower().replace(" ", "-"))
						if accuracy >= 0.4:
							T3SF_instance.inboxes_all[player] = channel
					
			json.dump(T3SF_instance.inboxes_all,open(f"inboxes_{T3SF_instance.platform}.json", "w"))

			for player in T3SF_instance.inboxes_all:
				mensaje_inboxes += f"Inbox {player} [{T3SF_instance.inboxes_all[player]}]\n"

			T3SF_instance.response_auto = await self.EditMessage(response=T3SF_instance.response_auto, color="YELLOW", title = f"📩 Inboxes fetched! [{len(T3SF_instance.inboxes_all)}]", description=mensaje_inboxes)
			
			T3SF_instance.regex_ready = True

		elif T3SF_instance.fetch_inboxes == True:

			T3SF_instance.response_auto = await self.SendMessage(channel = T3SF_instance._ctx['channel'], color="CYAN", title="💬 Fetching inboxes...", description=f"Please wait while we fetch all the inboxes in this server!")

			channels = await T3SF_instance.app.client.conversations_list(types="public_channel,private_channel")

			for channel in channels['channels']:
				T3SF_instance.ch_names_list.append(channel['name'])

			channels_itinerator = 0
			regex = ""
			past_channel = None

			while regex == "":
				for channel in T3SF_instance.ch_names_list:
					if channels_itinerator == 0 and channel == past_channel:
						past_channel = channel
						channels_itinerator += 1
						continue
					
					match_channel = utils.regex_finder(str(channel).lower() + "-" + str(past_channel).lower())

					if match_channel != False:
						for character in past_channel:
							if character in channel:
								regex += character
							else:
								break
						break
					else:
						channels_itinerator += 1
						past_channel = channel
						continue

			image = {"image_url":"https://i.ibb.co/34rTqMH/image.png", "name": "regex"}
			buttons = [{"text":"Yes!", "style": "primary", "value": regex, "action_id": "regex_yes"},{"text":"No.", "style": "danger", "value": "click_me_456", "action_id": "regex_no"}]
			
			T3SF_instance.response_auto = await self.EditMessage(response=T3SF_instance.response_auto, color="GREEN", title = "ℹ️ Regex detected!", description = f"Please confirm if the regex detected for the channels, is correct so we can get the inboxes!\n\nExample:\ninbox-legal\nThe regex should be `inbox-`\n\n*Detected regex:* `{regex}`\n\n\nPlease select your answer below.", image=image, buttons = buttons)
			T3SF_instance.regex_ready = False

		else:
			mensaje_inboxes = ""
			T3SF_instance.response_auto = await self.SendMessage(channel = T3SF_instance._ctx['channel'], color="CYAN", title="💬 Fetching inboxes...", description=f"Please wait while we fetch all the inboxes in this server!")

			for player in T3SF_instance.inboxes_all:
				mensaje_inboxes += f"Inbox {player} [{T3SF_instance.inboxes_all[player]}]\n"

			T3SF_instance.response_auto = await self.EditMessage(response=T3SF_instance.response_auto, color="YELLOW", title = f"📩 Inboxes fetched! [{len(T3SF_instance.inboxes_all)}]", description=mensaje_inboxes)

			T3SF_instance.regex_ready = True

	async def InjectHandler(self, T3SF_instance):
		image = None

		author = {"name": T3SF_instance._inject["From"], "date": T3SF_instance._inject["Date"]}

		player = T3SF_instance._inject['Player']

		if "Photo" in T3SF_instance._inject and T3SF_instance._inject['Photo'] != '':
			if "Picture Name" in T3SF_instance._inject and T3SF_instance._inject['Picture Name'] == '' or "Photo" not in T3SF_instance._inject:
				attachment_name = "attachment.jpg"
			else:
				attachment_name = T3SF_instance._inject['Picture Name']
			image = {"name": attachment_name, "image_url": T3SF_instance._inject['Photo']}
		
		if "Profile" in T3SF_instance._inject and T3SF_instance._inject['Profile'] != '':
			author["image_url"] = T3SF_instance._inject['Profile']
			
		else:
			profile_pic = random.choice([
				"https://ssl.gstatic.com/ui/v1/icons/mail/profile_mask2.png",
				"https://lh3.googleusercontent.com/-XdUIqdMkCWA/AAAAAAAAAAI/AAAAAAAAAAA/4252rscbv5M/photo.jpg",
				"https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcSTt8Dg9RL4IGOjsJ2Fr-lXThf-DGM5YgPB6j5rD8tHQ9RLrU-03H4dYeskL01FNajqL_0&usqp=CAU"
			])
			author["image_url"] = profile_pic
		
		await self.SendMessage(channel = T3SF_instance.inboxes_all[player], title=T3SF_instance._inject['Subject'], description=T3SF_instance._inject["Script"], image=image, author=author, color="CYAN")

		return True

	async def PollHandler(self, T3SF_instance):
		T3SF_instance.poll_answered = False
		
		image = None

		player = T3SF_instance._inject['Player']

		if "Photo" in T3SF_instance._inject and T3SF_instance._inject['Photo'] != '':
			if "Picture Name" in T3SF_instance._inject and T3SF_instance._inject['Picture Name'] == '' or "Photo" not in T3SF_instance._inject:
				attachment_name = "attachment.jpg"
			else:
				attachment_name = T3SF_instance._inject['Picture Name']
			image = {"name": attachment_name, "image_url": T3SF_instance._inject['Photo']}

		poll_options = T3SF_instance._inject['Poll'].split('|')
		
		actual_real_time = re.sub("([^0-9])", "", T3SF_instance._inject['Real Time'])[-2:]
		
		next_real_time = re.sub("([^0-9])", "", T3SF_instance.data[int(T3SF_instance._inject['#'])]['Real Time'])[-2:]

		diff = int(next_real_time) - int(actual_real_time)
		if diff < 0:
			diff_no_real = int(actual_real_time) - int(next_real_time)
			diff = 60 - diff_no_real 

		diff_secs = diff * 60

		description = T3SF_instance._inject["Script"] + f"\n\nYou have {diff} minute(s) to answer this poll!"

		buttons = [{"text": poll_options[0], "style": "primary", "value": 'option1', "action_id": "option1"},{"text":poll_options[1], "style": "primary", "value": "option2","action_id": "option2"}]

		T3SF_instance.response_poll = await self.SendMessage(channel = T3SF_instance.inboxes_all[player], title=T3SF_instance._inject['Subject'], description=description, image=image, buttons=buttons, color="YELLOW")

		return True

	async def PollAnswerHandler(self, T3SF_instance, body=None, payload=None):
		poll_msg = body['message']['attachments'][0]['fallback']
		poll_msg = poll_msg[: poll_msg.rfind('\n')]
		action_user = body['user']['username']
		selected_option = payload['text']['text']
		description = f'{poll_msg}\n\n@{action_user} selected: {selected_option}'

		T3SF_instance.poll_answered = True
		T3SF_instance.response_poll = await self.EditMessage(color = "GREEN", title="Poll Answered!", description=description, response=T3SF_instance.response_poll)
		await T3SF_instance.NotifyGameMasters(type_info="poll_answered", data={'msg_poll':poll_msg,'answer':selected_option,'user':action_user})
		return True

	async def SendMessage(self, T3SF_instance=None, color=None, title=None, description:str=None, channel=None, image=None, author=None, buttons=None, text_input=None, checkboxes=None):
		if channel == None:
			channel = self._ctx['channel']

		try:
			attachments = self.Formatter(title=title, description=description, color=color, image=image, author=author, buttons=buttons, text_input=text_input, checkboxes=checkboxes)
			self.response = await self.app.client.chat_postMessage(channel = channel, attachments = attachments)
			return self.response

		except Exception as e:
			raise

	async def EditMessage(self, T3SF_instance=None, color=None, title:str=None, description:str=None, response=None, image=None, author=None, buttons=None, text_input=None, checkboxes=None):
		self.response = await self.app.client.chat_update(channel=response['channel'], ts=response['ts'], attachments = self.Formatter(title=title, description=description, color=color, image=image, author=author, buttons=buttons, text_input=text_input, checkboxes=checkboxes))
		return self.response


from T3SF.gui.core import MessageAnnouncer
from datetime import datetime
import json
import uuid


class T3SF_Logger:
    @staticmethod
    def emit(message,message_type="DEBUG"):
        try:
            message = message.replace("\n"," ")

            # Create a dictionary with the message data
            message_data = {
                "id" : str(uuid.uuid4()),
                "type" : message_type,
                "content": message,
                "timestamp": datetime.now().strftime("%H:%M:%S")
            }

            # Convert the dictionary to a JSON string
            sse_msg = f"data: {json.dumps(message_data)}\n\n"

            announcer = MessageAnnouncer()
            announcer.announce(msg=sse_msg)

            # Do something with the EventSource-formatted message, such as writing it to a file
            with open('logs.txt', 'a+') as f:
                f.write(sse_msg)

            # Print some critical information to the terminal
            if message_type in ['WARN', 'ERROR']:
                red = '\033[91m\033[5m'
                yellow = '\033[93m\033[5m'
                nc = '\033[0m'

                if message_type == "WARN":
                    text = yellow + "[!] " + nc + message
                else:
                    text = red + "[✗] " + nc + message
                print(text)

        except Exception as e:
            print(f"We could not print this message on the webpage:\n{message}")


from difflib import SequenceMatcher
from collections import Counter
import os

def is_docker():
	"""
	Detects if the script is running inside a docker environment.
	"""
	path = '/proc/self/cgroup'
	return (
		os.path.exists('/.dockerenv') or
		os.path.isfile(path) and any('docker' in line for line in open(path))
	)

def similar(a, b):
	"""
	Based in graphics, find the similarity between 2 strings.
	"""
	return SequenceMatcher(None, a, b).ratio()

def regex_finder(input):
	"""
	Matches repeated words counting the 
	amount of times the word is being repeated.
	"""
	words = input.split('-')
	dict = Counter(words)
	for key in words:
		if dict[key]>1:
			return key
	return False


from flask import Flask, render_template, Response, stream_with_context, Blueprint, request
import os, signal, time
import threading
import logging
import uuid
import queue
from datetime import datetime
import json
from T3SF import utils

bp = Blueprint('t3sf', __name__)

platform = None
msel_file = None

class GUI(Flask):
	def __init__(self, platform_run, MSEL, import_name=__name__, *args, **kwargs):
		global platform, msel_file
		msel_file = MSEL
		platform = platform_run
		super().__init__(import_name, *args, **kwargs)
		self.config['SECRET_KEY'] = uuid.uuid4().hex
		self.register_blueprint(bp)
		self.start()

	def start_flask_app(self):
		log = logging.getLogger('werkzeug')
		log.setLevel(logging.WARNING)
		if utils.is_docker():
			print(" * GUI active on - http://0.0.0.0:5000")
			self.run(debug=False, threaded=True, use_reloader=False, host="0.0.0.0")
		else:
			print(" * GUI active on - http://127.0.0.1:5000")
			self.run(debug=False, threaded=True, use_reloader=False)

	def start(self):
		flask_app_thread = threading.Thread(target=self.start_flask_app)
		flask_app_thread.start()

listeners = [] 
class MessageAnnouncer:
	def __init__(self):
		self.listeners = []

	def listen(self):
		global listeners
		listeners.append(queue.Queue(maxsize=10))
		return listeners[-1]

	def announce(self, msg):
		global listeners
		# We go in reverse order because we might have to delete an element, which will shift the
		# indices backward
		for i in reversed(range(len(listeners))):
			try:
				listeners[i].put_nowait(msg)
			except queue.Full:
				del listeners[i]

# SSE 
@bp.route('/stream', methods=['GET'])
def listen():
	def stream():
		# Yield the previous logs
		try:
			with open('logs.txt', 'r+') as f:
				for line in f:
					yield line
		except FileNotFoundError:
			with open('logs.txt', 'a+') as f:
				for line in f:
					yield line

		messages = MessageAnnouncer().listen()  # returns a queue.Queue
		while True:
			msg = messages.get()  # blocks until a new message arrives
			yield msg

	return Response(stream(), mimetype='text/event-stream')

@bp.route('/stream_news', methods=['GET'])
def listen_news():
	def stream():
		messages = MessageAnnouncer().listen()  # returns a queue.Queue
		while True:
			msg = messages.get()  # blocks until a new message arrives
			yield msg

	return Response(stream(), mimetype='text/event-stream')

# Views 
@bp.route('/')
def index():
	return render_template('index.html', active_page="logs_viewer", platform=platform.lower())

@bp.route('/msel-viewer')
def msel_viewer():
	return render_template('msel_viewer.html', active_page="msel_viewer")

@bp.route('/env-creation')
def env_creation():
	from T3SF import T3SF
	T3SF = T3SF(platform=platform.lower())
	T3SF.IncidentsFetcher(MSEL=msel_file)

	return render_template('env_creation.html', active_page="env_creation", T3SF=T3SF)

# Actions
@bp.route('/start')
async def start_exercise():
	# Start the async task

	if platform.lower() == "discord":
		from T3SF.discord.bot import run_async_incidents
		try:
			server_id = int(request.args.get("server"))
			await run_async_incidents(server=server_id)

		except (ValueError, TypeError):
			msg = {
				"id": str(uuid.uuid4()),
				"type": "ERROR",
				"content": "The Token was not provided or is not an integer.",
				"timestamp": datetime.now().strftime("%H:%M:%S")
			}
			MessageAnnouncer().announce(msg=f"data: {json.dumps(msg)}\n\n")
			return Response("Failed, Token was not provided or is not an integer.", mimetype='text/plain')

	elif platform.lower() == "slack":
		from T3SF.slack.bot import run_async_incidents
		await run_async_incidents()

	# Return an immediate response
	return Response("Started", mimetype='text/plain')

@bp.route('/stop')
def stop_exercise():
	# Kill the script
	msg = {
		"id" : str(uuid.uuid4()),
		"type" : "INFO",
		"content": "Exiting...",
		"timestamp": datetime.now().strftime("%H:%M:%S")
	}
	MessageAnnouncer().announce(msg=f"data: {json.dumps(msg)}\n\n")
	time.sleep(1)
	os._exit(0)

	# Return a response that triggers the WSGI server to shutdown the application
	return 'Shutting down...'

@bp.route('/create')
async def create_env():
	if platform.lower() == "discord":
		from T3SF.discord.bot import create_environment
		try:
			server_id = int(request.args.get("server"))
			await create_environment(server=server_id)

		except (ValueError, TypeError):
			msg = {
				"id": str(uuid.uuid4()),
				"type": "ERROR",
				"content": "The Token was not provided or is not an integer.",
				"timestamp": datetime.now().strftime("%H:%M:%S")
			}
			MessageAnnouncer().announce(msg=f"data: {json.dumps(msg)}\n\n")
			return Response("Failed, Token was not provided or is not an integer.", mimetype='text/plain')

	elif platform.lower() == "slack":
		from T3SF.slack.bot import create_environment
		await create_environment()

	return "Created"

@bp.route('/clear')
def clear_logs():
	# Clear the logs
	open('logs.txt', 'w').close()

	# Return a response
	return 'Logs cleared...'


from slack_bolt.adapter.socket_mode.async_handler import AsyncSocketModeHandler
from slack_bolt.app.async_app import AsyncApp
from slack_sdk.errors import SlackApiError
from dotenv import load_dotenv
from .slack import Slack
import warnings
import asyncio
import os, re
import json
import T3SF

load_dotenv()

# Disable the UserWarning message
warnings.filterwarnings("ignore", category=UserWarning)

app_slack = None
T3SF_instance = None
config_MSEL = None


class create_bot():
	def __init__(self, MSEL=None):
		global config_MSEL
		config_MSEL = MSEL
		pass

	async def slack_main(self):
		global app_slack
		app_slack = self.app_slack = AsyncApp(token=os.environ["SLACK_BOT_TOKEN"])

		handler = AsyncSocketModeHandler(self.app_slack, os.environ["SLACK_APP_TOKEN"])

		@app_slack.action(re.compile("regex"))
		async def regex_handler(ack, body, payload):
			await T3SF.T3SF.RegexHandler(self=T3SF_instance,ack=ack,body=body,payload=payload)

		@app_slack.action(re.compile("option"))
		async def poll_handler(ack, body, payload):
			await T3SF.T3SF.PollAnswerHandler(self=T3SF_instance,ack=ack,body=body,payload=payload)

		@app_slack.message("!start")
		async def start(message, say):
			"""
			Retrieves the !start command and starts to fetch and send the incidents.
			"""
			try:
				global T3SF_instance
				T3SF_instance = T3SF.T3SF(platform="slack", app=app_slack)
				await T3SF_instance.ProcessIncidents(MSEL = config_MSEL, function_type = "start", ctx=message)

			except Exception as e:
				print("ERROR - Start function")
				print(e)
				raise

		@app_slack.message('!ping')
		async def ping(message, say):
			"""
			Retrieves the !ping command and replies with this message.
			"""
			description = """
			PING localhost (127.0.0.1): 56 data bytes\n64 bytes from 127.0.0.1: icmp_seq=0 ttl=113 time=37.758 ms\n64 bytes from 127.0.0.1: icmp_seq=1 ttl=113 time=50.650 ms\n64 bytes from 127.0.0.1: icmp_seq=2 ttl=113 time=42.493 ms\n64 bytes from 127.0.0.1: icmp_seq=3 ttl=113 time=37.637 ms\n--- localhost ping statistics ---\n4 packets transmitted, 4 packets received, 0.0% packet loss\nround-trip min/avg/max/stddev = 37.637/42.135/50.650/5.292 ms\n\n_This is not real xD_"""
			await app_slack.client.chat_postMessage(channel = message['channel'], attachments = Slack.Formatter(color="GREEN", title="🏓 Pong!", description=description))

		@app_slack.event("message")
		async def not_interesting_messages(body, logger):
			pass

		await handler.start_async()

async def start_bot():
	task1 = asyncio.create_task(create_bot(config_MSEL).slack_main())
	T3SF.T3SF_Logger.emit(message=f'Slack Bot is ready!', message_type="DEBUG")
	await asyncio.gather(task1)

async def run_async_incidents():
	global T3SF_instance
	T3SF_instance = T3SF.T3SF(platform="slack", app=app_slack)
	await T3SF_instance.ProcessIncidents(MSEL = config_MSEL, function_type = "start", ctx={"channel":"gm-chat"})

async def create_environment():
	global T3SF_instance

	T3SF_instance = T3SF.T3SF(platform="slack", app=app_slack)
	areas_msel = T3SF_instance.IncidentsFetcher(MSEL = config_MSEL)
	
	admins_users = await get_admins()

	await create_gm_channels(admins=admins_users)

	try:
		inbox_name = "inbox"
		extra_chnls = ['chat','decision-log']
		players_list_local = areas_msel

		channels_ids = []

		for player in players_list_local:
			inbox =  inbox_name + "-" + sanitize_channel_name(player)
			try:
				channel_id_inbox = await create_channel_if_not_exists(channel_name=inbox, private=True)
				
				if channel_id_inbox:
					await app_slack.client.conversations_invite(channel=channel_id_inbox, users=admins_users)
				
					channels_ids.append(channel_id_inbox)

					T3SF.T3SF_Logger.emit(message=f'Player [{player}] - Channel {inbox_name} created', message_type="DEBUG")

				for extra in extra_chnls:
					channel = extra + "-" + sanitize_channel_name(player)
				
					channel_id_extra = await create_channel_if_not_exists(channel_name=channel, private=True)
					if channel_id_extra:
						await app_slack.client.conversations_invite(channel=channel_id_extra, users=admins_users)
				
						channels_ids.append(channel_id_extra)

						T3SF.T3SF_Logger.emit(message=f'Player [{player}] - Channel {channel} created', message_type="DEBUG")

			except Exception as e:
				print("ERROR - Create function")
				print(e)
				pass
		
		T3SF.T3SF_Logger.emit(message=f"Player's Channels created!", message_type="INFO")
		print(channels_ids)
		T3SF.T3SF_Logger.emit(message=f'Channels IDs: {channels_ids}', message_type="DEBUG")

	except Exception as e:
		print("ERROR - Create function")
		print(e)
		raise

async def get_admins():
	admins_users = []
	try:
		response = await app_slack.client.users_list()
		# Loop through the list of users and check if they're an admin
		for user in response["members"]:
			try:
				if user["is_admin"]:
					# Store the user ID of the workspace admin
					admins_users.append(user['id'])
			except KeyError:
				pass
		return admins_users

	except SlackApiError as e:
		if e.response["error"] == "ratelimited":
			T3SF.T3SF_Logger.emit(message=f'Slack is rate limiting us. Please wait a few seconds and try again.', message_type="ERROR")

	except Exception as e:
		print("ERROR - Get Admins function")
		print(e)
		raise
	
async def create_channel_if_not_exists(channel_name, private=True):
	try:
		# Call the conversations_list method using the app client
		response = await app_slack.client.conversations_list()

		# Check if a channel with the given name already exists
		channel_id = None
		for channel in response["channels"]:
			if channel["name"] == channel_name:
				channel_id = channel["id"]
				break

		# If the channel doesn't exist, create it using the conversations_create method
		if not channel_id:
			response = await app_slack.client.conversations_create(name=channel_name ,is_private=private)
			channel_id = response["channel"]["id"]

		# Return the ID of the channel
		return channel_id

	except SlackApiError as e:
		if e.response["error"] == "name_taken":
			T3SF.T3SF_Logger.emit(message=f'Channel {channel_name} already taken.', message_type="WARN")
		else:
			print(f"Error: {e}")
		return False

async def create_gm_channels(admins):
	channels = ['gm-chat','gm-logs']
	for channel in channels:
		channel_id = await create_channel_if_not_exists(channel)
		if channel_id:
			await app_slack.client.conversations_invite(channel=channel_id, users=admins)

	T3SF.T3SF_Logger.emit(message=f'Game Master channels created.', message_type="INFO")

	return True

def sanitize_channel_name(name):
	# Replace whitespace with a hyphen
	sanitized_name = re.sub(r"\s", "-", name)

	# Remove special characters except for hyphens and periods
	sanitized_name = re.sub(r"[^\w.-]", "-", sanitized_name)

	# Replace consecutive hyphens and periods with a single hyphen
	sanitized_name = re.sub(r"[-.]+", "-", sanitized_name)

	# Remove leading and trailing hyphens
	sanitized_name = sanitized_name.strip("-")

	# Convert to lowercase
	sanitized_name = sanitized_name.lower()

	# Truncate to a maximum length of 80 characters
	sanitized_name = sanitized_name[:80]

	return sanitized_name


#!/usr/bin/python3

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="T3SF",
    version="2.1-rc1",
    author="BASE4 Security",
    author_email="jlanfranconi@base4sec.com",
    description="Technical Tabletop Exercises Simulation Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Base4Security/T3SF",
    project_urls={
        "Bug Tracker": "https://github.com/Base4Security/T3SF/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    include_package_data=True,
    install_requires=[
        'Flask[Async]',
        'python-dotenv'
    ],
    extras_require={
        'Slack': ['slack_bolt', 'aiohttp'],
        'Discord': ['discord.py']
    },
)



import importlib
import asyncio
import signal
import json
import time
import re
import os

# Internal logging
from .logger import T3SF_Logger
# / Internal logging

def keyboard_interrupt_handler(signal_num, frame):
	T3SF_Logger.emit(message=f"KeyboardInterrupt (ID: {signal_num}) has been caught. Exiting...", message_type="WARN")
	os._exit(1)

# Associate the signal handler function with SIGINT (keyboard interrupt)
signal.signal(signal.SIGINT, keyboard_interrupt_handler)


class T3SF(object):
	def __init__(self, bot = None, app = None, platform = None):
		self.response = None
		self.process_wait = False
		self.process_quit = False
		self.regex_ready = None
		self.incidents_running = False
		self.poll_answered = False
		self.ch_names_list = []
		self.players_list = []
		self.platform = platform.lower()
		self.guild_id = None

		try:
			if os.path.isfile(os.getcwd() + f"/inboxes_{self.platform}.json"):
				self.inboxes_all = json.load(open(f"inboxes_{self.platform}.json", encoding='utf-8-sig'))
				self.fetch_inboxes = False

				T3SF_Logger.emit(message="Locally retrieved Inboxes", message_type="DEBUG")
			else:
				self.fetch_inboxes = True 
				self.inboxes_all = {}

		except Exception:
			self.fetch_inboxes = True
			self.inboxes_all = {}

		if self.platform == "discord":
			from .discord import Discord
			self.bot = bot
			self.discord = Discord(bot)

		elif self.platform == "slack":
			from .slack import Slack
			self.app = app
			self.slack = Slack(app)

	async def TimeDifference(self, actual_real_time, previous_real_time, itinerator=int, resumed=bool):
		"""
		This function is used to Get the difference between two injects. It will make the bot sleep and inform the Game Masters.
		"""
		try:
			self.diff = int(actual_real_time) - int(previous_real_time)
			
			if self.diff < 0:
				self.diff_no_real = int(previous_real_time) - int(actual_real_time)
				self.diff = 60 - self.diff_no_real 
			
			self.diff_secs = self.diff * 60
			
			if resumed != True:     
				itinerator = itinerator + 1

			description = f"The bot is Up and running!\n\nIncident: {itinerator}/{len(self.data)}\n\nWaiting {self.diff} minute(s) ({self.diff_secs} sec.) to send it."

			# await self.EditMessage(style="simple", color = "CYAN", title="⚙️ Bot running...", description=description, response=self.msg_gm)

			T3SF_Logger.emit(f'We have a difference of {self.diff} minute(s) - {self.diff_secs} seconds', message_type="INFO")
			
			await asyncio.sleep(self.diff_secs)

			if "Poll" in self._inject and self._inject['Poll'] != '' and self.poll_answered == False:
				description = self._inject["Script"] + f"\n\n@channel Poll not answered within {self.diff} minute(s), Time's Up!"
				await self.EditMessage(style="custom", variable="T3SF_instance.response_poll", color = "RED", title="Poll time ended!", description=description, response=self.response_poll)
				await self.NotifyGameMasters(type_info="poll_unanswered", data={'msg_poll':self._inject["Script"]}) 

		except Exception as e:
			print("Get Time Difference")
			print(e)
			T3SF_Logger.emit("Get Time Difference", message_type="ERROR")
			T3SF_Logger.emit(e, message_type="ERROR")
			raise

	def IncidentsFetcher(self, MSEL:str):
		"""
		Retrieves the incidents from the desired source, chosen in the config file.
		"""
		T3SF_Logger.emit(message="Reading MSEL", message_type="DEBUG")
		if MSEL:
			self.data = json.load(open(MSEL, encoding='utf-8-sig'))

			for inject in self.data:
				player = inject['Player']
				if player not in self.players_list:
					self.players_list.append(player)
			T3SF_Logger.emit(message="We have the inboxes right now", message_type="DEBUG")
			T3SF_Logger.emit(message="Incidents ready", message_type="DEBUG")
			return self.players_list
		else:
			raise RuntimeError("Please set a method to retrieve the TTXs with the argument `MSEL` inside the `start` function.")

	async def start(MSEL:str, platform, gui=False):
		if gui == True:
			T3SF_Logger.emit(message="Starting GUI", message_type="DEBUG")
			gui_module = importlib.import_module("T3SF.gui.core")
			gui_module.GUI(platform_run=platform, MSEL=MSEL)
		
		if platform.lower() == "slack":
			bot_module = importlib.import_module("T3SF.slack.bot")

		elif platform.lower() == "discord":
			bot_module = importlib.import_module("T3SF.discord.bot")

		else:
			raise ValueError("Invalid platform")

		T3SF_Logger.emit(message="Starting BOT", message_type="DEBUG")
		bot_module.create_bot(MSEL=MSEL)
		await bot_module.start_bot()

	async def NotifyGameMasters(self, type_info=str, data=None):
		"""
		Notify the Game Masters of the different states of the bot, through messages.
		"""
		try:
			if type_info == "start_normal":
				title = "⚙ Starting bot..."
				description = "The bot it's heating up!\n\nGive us just a second!!"
				# self.msg_gm = await self.SendMessage(title = title, description = description, color="YELLOW")

			elif type_info == "started_normal":
				title = "Bot succesfully started! 🎈"
				description = "The bot is Up and running!\n\nLets the game begin!!"
				# self.msg_gm = await self.EditMessage(title = title, description = description, color="GREEN", response=self.msg_gm)
			
			elif type_info == "start_resumed":
				title = "⚙ Resuming bot..."
				description = "The bot it's trying to resume from the desired point!\n\nGive us just a few seconds!!"
				# self.msg_gm = await self.SendMessage(title = title, description = description, color="YELLOW")
			
			elif type_info == "started_resumed":
				title = "Bot succesfully started! 🎈"
				description = "The bot is Up and running!\n\nLets the game begin!!"
				# self.msg_gm = await self.EditMessage(title = title, description = description, color="GREEN", response=self.msg_gm)
			
			elif type_info == "finished_incidents":
				title = "🎉 Bot Finished succesfully! 🎉"
				description = "The bot've just completed the entire game!\n\nHope to see you again soon!!"
				# self.msg_gm = await self.EditMessage(title = title, description = description, color="GREEN", response=self.msg_gm)

			elif type_info == "poll_answered":
				title = "📊 Poll Answered"
				description = f"Poll Question: {data['msg_poll']}\nSelected Answer: {data['answer']}\nBy: @{data['user']}"
				# await self.SendMessage(title = title, description = description, color="GREEN", unique=True)

			elif type_info == "poll_unanswered":
				title = "📊 Poll Not Answered"
				description = f"Poll Question: {data['msg_poll']}\nNot answered by anyone."
				# await self.SendMessage(title = title, description = description, color="RED", unique=True)

			T3SF_Logger.emit(message=description, message_type="INFO")
			return True

		except Exception as e:
			print("NotifyGameMasters")
			print(e)
			T3SF_Logger.emit("NotifyGameMasters", message_type="ERROR")
			T3SF_Logger.emit(e, message_type="ERROR")
			raise

	async def ProcessIncidents(self, MSEL:str, ctx, function_type:str=None, itinerator:int=0):
		"""
		Process the incidents from the MSEL file.
		"""
		self.IncidentsFetcher(MSEL)
		try:
			self._ctx = ctx

			await self.InboxesAuto()

			while self.regex_ready == False:
				await asyncio.sleep(2)

			if function_type == "start":
				await self.NotifyGameMasters(type_info="start_normal")  # Sends a message regarding the bot's start procedure.
			else:
				await self.NotifyGameMasters(type_info="start_resumed")  # Sends a message regarding the bot's restart procedure.
			
			bypass_time = True

			for information in self.data:
				if self.process_quit == True:
					break
				
				if self.process_wait == True:
					while self.process_wait == True:
						await asyncio.sleep(5)

				if itinerator == 0: # Set a variable to get the actual timestamp and the past one, after that checks for differences.
					itinerator_loop = itinerator
				else:
					if function_type == "resume":
						itinerator_loop = itinerator - 2
					else:
						itinerator_loop = itinerator - 1
				
				if int(information["#"]) != itinerator and function_type == "resume": # Checks if the incident ID is the same as the desired starting point.
					pass
				
				else:
					actual_real_time = re.sub("([^0-9])", "", information["Real Time"])[-2:]
					
					previous_real_time = re.sub("([^0-9])", "", self.data[itinerator_loop]["Real Time"])[-2:]
					
					T3SF_Logger.emit(f"Previous {previous_real_time} - Actual {actual_real_time}", message_type="DEBUG")

					if previous_real_time != actual_real_time and function_type == "start":
						await self.TimeDifference(actual_real_time, previous_real_time, resumed=False, itinerator=itinerator) # Check the amount of seconds between both timestamps.

					elif previous_real_time != actual_real_time and bypass_time != True:
						await self.TimeDifference(actual_real_time, previous_real_time, resumed=True, itinerator=itinerator) # Check the amount of seconds between both timestamps.

					T3SF_Logger.emit(f'Inject {information["#"]}/{len(self.data)}', message_type="INFO")

					if "Poll" in information and information['Poll'] != '':
						await self.SendPoll(inject = information)
						
					else:
						await self.SendIncident(inject = information) # Sends the incident to the desired chats.

					if function_type == "start":
						if itinerator == 0:
							await self.NotifyGameMasters(type_info="started_normal") # Informs that the bot succesfully started.
					else:
						if bypass_time == True:
							await self.NotifyGameMasters(type_info="started_resumed") # Informs that the bot succesfully restarted.
							bypass_time = False

					itinerator += 1
			
			await self.NotifyGameMasters(type_info="finished_incidents") # Informs that the script is completed and there's no remaining incidents.
			self.process_quit = False
			self.process_wait = False
			self.incidents_running = False

		except Exception as e:
			print("ProcessIncidents function")
			print(e)
			T3SF_Logger.emit("ProcessIncidents function", message_type="ERROR")
			T3SF_Logger.emit(e, message_type="ERROR")
			raise

	async def SendIncident(self, inject):
		try:
			self._inject = inject

			if self.platform == "discord":
				await self.discord.InjectHandler(T3SF_instance=self)

			elif self.platform == "slack":
				await self.slack.InjectHandler(T3SF_instance=self)

			return True

		except Exception as e:
			print("SendIncident")
			print(e)
			T3SF_Logger.emit("SendIncident", message_type="ERROR")
			T3SF_Logger.emit(e, message_type="ERROR")
			raise

	async def SendPoll(self, inject):
		try:
			self._inject = inject

			if self.platform == "discord":
				await self.discord.PollHandler(T3SF_instance=self)

			elif self.platform == "slack":
				await self.slack.PollHandler(T3SF_instance=self)

			return True

		except Exception as e:
			print("SendPoll")
			print(e)
			T3SF_Logger.emit("SendPoll", message_type="ERROR")
			T3SF_Logger.emit(e, message_type="ERROR")
			raise

	async def InboxesAuto(self, message=None):
		if self.platform == "discord":
			await self.discord.InboxesAuto(T3SF_instance=self)

		elif self.platform == "slack":
			await self.slack.InboxesAuto(T3SF_instance=self, regex=None)
	
	async def RegexHandler(self, ack=None, body=None, payload=None, inbox=None):
		if self.platform == "slack":
			await ack()
			text_input = None
			image = None
			regex = None

			if payload['action_id'] == "regex_yes":
				regex = body['actions'][0]['value']
				color="GREEN"
				title = "✨ Regex detected succesfully! ✨"
				description = f"Thanks for confirming the regex detected for the channels (I'm going to tell my creator he is so good coding :D ), we are going to use `{regex}` to match the inboxes"

			elif payload['action_id'] == "regex_no":
				color="RED"
				title = "ℹ️ Regex needed!"
				description = "Got it!\n Unluckily, but here we go...\nPlease send me the regex for the channels, so we can get the inboxes!\n\nExample:\ninbox-legal\nThe regex should be `inbox-`"
				text_input = {"action_id": "regex_custom", "label": "Please type the desired regex. EG: inbox-", "dispatch_action": True}
				image = {"image_url":"https://i.ibb.co/34rTqMH/image.png", "name": "regex"}

			elif payload['action_id'] == "regex_custom":
				regex = body['actions'][0]['value']
				color="GREEN"
				title="✅ Regex accepted!"
				description=f"Thanks for confirming the regex for the channels, we are going to use `{regex}` to match the inboxes!"

			self.response_auto = await self.EditMessage(title = title, description = description, color=color, image=image, text_input=text_input, response=self.response_auto)

			if regex != None:
				await self.slack.InboxesAuto(T3SF_instance=self, regex=regex)

	async def PollAnswerHandler(self, ack=None, body=None, payload=None, query=None):
		if self.platform == "discord":
			await self.discord.PollAnswerHandler(T3SF_instance=self, interaction=payload)
			return True

		elif self.platform == "slack":
			await ack()
			await self.slack.PollAnswerHandler(T3SF_instance=self, body=body, payload=payload)
			return True

	async def SendMessage(self,
		color=None, 
		title:str=None, 
		description:str=None, 
		channel=None, 
		image=None, 
		author=None, 
		buttons=None, 
		text_input=None, 
		checkboxes=None,
		view=None,
		unique=False,
		reply_markup=None):

		if self.platform == "discord":
			if unique == True:
				self.gm_poll_msg = await self.discord.SendMessage(T3SF_instance=self, color=color, title=title, description=description, view=view, unique=unique)
				return self.gm_poll_msg
			else:
				self.response = await self.discord.SendMessage(T3SF_instance=self, color=color, title=title, description=description, view=view)
				return self.response

		elif self.platform == "slack":
			self.response = await Slack.SendMessage(self=self.slack, channel = channel, title=title, description=description, color=color, image=image, author=author, buttons=buttons, text_input=text_input, checkboxes=checkboxes)
			return self.response

	async def EditMessage(self,
		color=None, 
		style:str="simple", 
		title:str=None, 
		description:str=None,
		response=None, 
		variable=None, 
		image=None, 
		author=None, 
		buttons=None, 
		text_input=None, 
		checkboxes=None,
		view=None,
		reply_markup=None):

		if self.platform == "discord":
			if style == "simple":
				self.response = await self.discord.EditMessage(T3SF_instance=self, color=color, title=title, description=description, view=view)
				return self.response
			else:
				self.response = await self.discord.EditMessage(T3SF_instance=self, color=color, title=title, description=description, view=view, variable=variable, style="custom")
				return self.response

		elif self.platform == "slack":
			self.response = await self.slack.EditMessage(response=response, title=title, description=description, color=color, image=image, author=author, buttons=buttons, text_input=text_input, checkboxes=checkboxes)
			return self.response